==> btrview-0.4.0-py3-none-any.whl: btrview/subvolume.py <==
"""Subvolume Classes and errors."""
import subprocess
import re
from pathlib import Path, PurePath
from datetime import datetime
from typing import Self
from dataclasses import dataclass

from btrview.utils import get_UUIDs, run

class NotASubvolumeError(NotADirectoryError):
    """Throw when a directory isn't a subvolume"""

@dataclass(frozen=True)
class Mount:
    """Basic class for working with mounted subvolumes."""
    fsroot: PurePath
    target: Path

    def resolve(self, path: str) -> Path:
        """Returns the resolved path of another path"""
        fsroot_str = str(self.fsroot)
        target_str = str(self.target)
        if fsroot_str == "/":
            target_str = target_str + "/"
        path_str = str(path)
        new_path = path_str.replace(fsroot_str,target_str,1).replace("//","/",1)
        return Path(new_path)

    def __str__(self) -> str:
        return f"{self.fsroot} on {self.target}"

class Subvolume:
    """Class representing a btrfs subvolume"""
    def __init__(self, props: dict[str,str|None], mounts: tuple[Mount, ...],
                 deleted: bool = False, show: bool = False) -> None:
        self.props = props
        self.mounts = mounts
        self.deleted = deleted
        self._show = show

    @property
    def paths(self) -> list[Path]:
        if not self["btrfs Path"]:
            return []
        btr_path = Path(self["btrfs Path"])
        return [mount.resolve(btr_path) for mount in self.mounts if btr_path.is_relative_to(mount.fsroot)]

    @property
    def mounted(self) -> bool:
        return bool(self.paths)

    @property
    def mount_points(self) -> tuple[Path, ...]:
        targets = [mount.target for mount in self.mounts]
        return tuple(path for path in self.paths if path in targets)

    def parent(self, p_type: str) -> str | None:
        """Returns parent UUID or ID string"""
        match p_type:
            case "snap":
                parent = self["Received UUID"] or self["Parent UUID"]
            case "subvol":
                parent = self["Parent ID"]
            case _:
                parent = None
        return parent

    def id(self, p_type: str) -> str | None:
        """Returns subvolume UUID or ID string"""
        match p_type:
            case "snap":
                ID = self["UUID"]
            case "subvol":
                ID = self["Subvolume ID"]
            case _:
                ID = None
        return ID

    @classmethod
    def from_UUID(cls, uuid: str, path: str | Path, mounts: tuple[Mount, ...]) -> Self:
        """Creates subvolume from the subvolumes UUID and any path on the filesystem"""
        cmd = f"btrfs subvolume show -u {uuid} {path}"
        props = cls._run_cmd(cmd)
        return cls(props, mounts, show = True)

    @classmethod
    def from_ID(cls, ID: str, path: str | Path, mounts: tuple[Mount, ...]) -> Self:
        """Creates subvolume from subvolume's ID and any path on the filesystem"""
        cmd = f"btrfs subvolume show -r {ID} {path}"
        props = cls._run_cmd(cmd)
        return cls(props, mounts, show = True)

    @classmethod
    def _run_cmd(cls, cmd: str) -> dict[str, str | None]:
        """Runs the shell command and returns the prop dictionary
        if the command doesn't error"""
        out = run(cmd)
        if out.returncode != 0:
            raise NotASubvolumeError
        props = cls._get_props(out.stdout)
        return props

    @classmethod
    def _get_props(cls, btrfs_show_text: str) -> dict[str, str | None]:
        """Creates btrfs prop dict based on the output of 
        btrfs subvolume show."""
        subvol = {}
        lines = btrfs_show_text.splitlines()
        subvol["btrfs Path"] = ("/" + lines[0]).replace("//","/")
        for line in lines[1:]:
            if re.search(r":\s+",line):
                k,v = line.split(":",maxsplit=1)
                k = k.strip()
                v = v.strip()
                v = None if v == "-" else v
                subvol[k] = v
        return subvol

    @staticmethod
    def is_btrfs(path: Path) -> bool:
        """Returns true if path is part of a btrfs filesystem."""
        response = run(f"btrfs filesystem usage '{path}'")
        return response.returncode == 0

    def __getitem__(self, key: str) -> str | None:
        """Returns the item from the props dictionary, but instead
        of throwing a key error, returns None"""
        if key in self.props:
            return self.props.get(key)
        elif not self._show and not self.deleted:
            cmd = f"btrfs subvolume show -u {self['UUID']} {self.mounts[0].target}"
            props = self._run_cmd(cmd)
            self.props |= props
            self._show = True
        return self.props.get(key)

    def __str__(self) -> str:
        string = self["Name"] or str(self["UUID"])
        if mps := self.mount_points:
            mp_string = ", ".join(str(mp) for mp in mps)
            string = f"{string} on: {mp_string}"
        return string

    def __hash__(self) -> int:
        return hash(self["UUID"])

    def __eq__(self, other) -> bool:
        return self["UUID"] == other["UUID"]

class MountedSubvolume(Subvolume):
    """Class representing a mounted subvolume. Differs from a normal Subvolume
    in that it can be snapshotted and sent since there's a path to it."""
    def __init__(self, props: dict[str, str| None], path: Path) -> None:
        self.props = props
        self.path = path

    @classmethod
    def from_path(cls, path: str | Path) -> Self:
        cmd = f"btrfs subvolume show {path}"
        props = cls._run_cmd(cmd)
        path = Path(path)
        return cls(props, path)

    def same_mount(self, path: Path) -> bool:
        """Returns true if a subvolume is on the same filesystem as a specified path"""
        return get_UUIDs(self.path) == get_UUIDs(path)

    def snapshot(self, snap_dir: Path, format_str: str = "") -> Self:
        """Take a snapshot of the subvolume and save it to snap_dir"""
        if not snap_dir.is_dir():
            raise NotADirectoryError
        if format_str:
            timestamp = datetime.now().astimezone().strftime(format_str)
        else:
            timestamp = datetime.now().astimezone().isoformat()
        snap_file = snap_dir / timestamp
        out = run(f"btrfs subvolume snapshot -r '{self.path}' '{snap_file}'")
        return type(self).from_path(snap_file)

    def send(self, path: Path) -> Self:
        """Sends the subvolume to another filesystem"""
        try:
            #note that `shell=True` has some security implications with
            #user input, but figuring out how to properly do this with
            #subprocess.Popen() is not what I want to do right now.
            subprocess.run(f"btrfs send {self} | btrfs receive {path}", shell=True, check=True)
        except Exception as e:
            run(f"btrfs subvolume delete '{path}'")
            print(e)
        return type(self).from_path(path / self.path.name)

    def delete(self) -> Subvolume:
        """Deletes the subvolume"""
        run(f"btrfs subvolume delete '{self}'")
        props = {"UUID":self["UUID"]}
        return Subvolume(props, tuple(), deleted=True)

==> btrview-0.6.0-py3-none-any.whl: btrview/subvolume.py <==
"""Subvolume Classes and errors."""
from pathlib import Path, PurePath
from typing import Self
from dataclasses import dataclass

from btrview.utils import run
from btrview.btr_dict import BtrDict, BtrfsDict

class NotASubvolumeError(NotADirectoryError):
    """Throw when a directory isn't a subvolume"""

@dataclass(frozen=True)
class Mount:
    """Basic class for working with mounted subvolumes."""
    fsroot: PurePath
    target: Path

    def resolve(self, path: Path) -> Path:
        """Returns the resolved path of another path"""
        fsroot_str = str(self.fsroot)
        target_str = str(self.target)
        if fsroot_str == "/":
            target_str = target_str + "/"
        path_str = str(path)
        new_path = path_str.replace(fsroot_str,target_str,1).replace("//","/",1)
        return Path(new_path)

    def __str__(self) -> str:
        return f"{self.fsroot} on {self.target}"

class Subvolume:
    """Class representing a btrfs subvolume"""
    def __init__(self, props: BtrDict, mounts: tuple[Mount, ...],
                 deleted: bool = False, show: bool = False) -> None:
        self.props = props
        self.mounts = mounts
        self.deleted = deleted
        self._show = show

    @property
    def paths(self) -> list[Path]:
        """Returns all the reachable paths of the Subvolume"""
        if self["btrfs Path"] is None:
            return []
        btr_path = Path(self["btrfs Path"])
        paths = [mount.resolve(btr_path) for mount in self.mounts if btr_path.is_relative_to(mount.fsroot)]
        paths = [path for path in paths if path.exists()]
        return paths

    @property
    def mounted(self) -> bool:
        """Returns whether the subvolume is reachable via the filesystem"""
        return bool(self.paths)

    @property
    def root_subvolume(self) -> bool:
        """Returns whether the subvolume is the root_subvolume"""
        return self["Subvolume ID"] == "5"

    @property
    def snapshot(self) -> bool:
        """Returns whether the subvolume is a snapshot"""
        return bool(self["Parent UUID"])

    @property
    def mount_points(self) -> tuple[Path, ...]:
        """Returns mount points of Subvolume"""
        targets = [mount.target for mount in self.mounts]
        return tuple(path for path in self.paths if path in targets)

    def parent(self, p_type: str) -> str | None:
        """Returns parent UUID or ID string"""
        match p_type:
            case "snap":
                parent = self["Received UUID"] or self["Parent UUID"]
            case "subvol":
                parent = self["Parent ID"]
            case _:
                parent = None
        return parent

    def id(self, p_type: str) -> str:
        """Returns subvolume UUID or ID string"""
        match p_type:
            case "snap":
                ID = self["UUID"]
            case "subvol":
                ID = self["Subvolume ID"]
            case _:
                ID = self["UUID"]
        return ID

    @classmethod
    def from_UUID(cls, uuid: str, path: str | Path, mounts: tuple[Mount, ...]) -> Self:
        """Creates subvolume from the subvolumes UUID and any path on the filesystem"""
        cmd = f"btrfs subvolume show -u {uuid} {path}"
        props = cls._run_cmd(cmd)
        return cls(props, mounts, show = True)

    @classmethod
    def from_ID(cls, ID: str, path: str | Path, mounts: tuple[Mount, ...]) -> Self:
        """Creates subvolume from subvolume's ID and any path on the filesystem"""
        cmd = f"btrfs subvolume show -r {ID} {path}"
        props = cls._run_cmd(cmd)
        return cls(props, mounts, show = True)

    @classmethod
    def from_list(cls, list_str: str, mounts: tuple[Mount, ...]) -> Self:
        props = BtrfsDict.from_list(list_str).btr_dict
        return cls(props, mounts)

    @classmethod
    def from_deleted(cls, UUID: str) -> Self:
        """Creates subvolume from subvolume's ID and any path on the filesystem"""
        props: BtrDict = {"UUID":UUID,"Subvolume ID":UUID, "Name":UUID}
        return cls(props, tuple(), deleted=True)

    @classmethod
    def _run_cmd(cls, cmd: str) -> BtrDict:
        """Runs the shell command and returns the prop dictionary
        if the command doesn't error"""
        out = run(cmd)
        if out.returncode != 0:
            raise NotASubvolumeError
        props = BtrfsDict.from_show(out.stdout).btr_dict
        return props

    @staticmethod
    def is_btrfs(path: Path) -> bool:
        """Returns true if path is part of a btrfs filesystem."""
        response = run(f"btrfs filesystem usage '{path}'")
        return response.returncode == 0

    def __getitem__(self, key: str) -> str | None:
        """Returns the item from the props dictionary, but instead
        of throwing a key error, returns None"""
        if key in self.props:
            return self.props[key]
        elif self.deleted:
            #just assume it's None for deleted subvols. #could cause problems
            #in that deleted Subvolumes won't throw KeyError
            return None
        elif not self._show:
            cmd = f"btrfs subvolume show -u {self['UUID']} {self.mounts[0].target}"
            props = self._run_cmd(cmd)
            self.props |= props
            self._show = True
        try:
            return self.props[key]
        except KeyError:
            pass
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"Subvolume has no attribute or key '{key}'")

    def __str__(self) -> str:
        string = self["Name"]
        if mps := self.mount_points:
            mp_string = ", ".join(str(mp) for mp in mps)
            string = f"{string} on: {mp_string}"
        return str(string)

    def __hash__(self) -> int:
        return hash(self["UUID"])

    def __eq__(self, other) -> bool:
        return self["UUID"] == other["UUID"]
